/*
 * Copyright (c) 2010 RWTH Aachen University
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Hendrik vom Lehn <vomlehn@cs.rwth-aachen.de>
 */

#include <linux/module.h>
#include <linux/version.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/in.h>
#include <linux/inet.h>
#include <linux/errno.h>
#include <linux/kthread.h>
#include <linux/workqueue.h>
#include <linux/time.h>
#include <net/sock.h>

#include "wifi-emu.h"
#include "wifi-emu-msg.h"

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Hendrik vom Lehn <vomlehn@cs.rwth-aachen.de>");
MODULE_DESCRIPTION("Wifi Emulation Driver (UDP communication part)");

/*         */
/* defines */
/*         */
#define DEBUG_TIME 0

/*                                 */
/* function prototype declarations */
/*                                 */
static int com_thread(void *data); // receiver thread
static void wifi_emu_send_msg(uint8_t *msg, int len); // called by wifi-emu, queues work

// sends the queued packet
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20))
static void send_msg_wq(struct work_struct *work);
#else
static void send_msg_wq(void *data);
#endif

// init and cleanup
static int __init wifi_emu_udp_init(void);
static void __exit wifi_emu_udp_cleanup(void);
// free resources
static inline void free_resources(void);

/*                             */
/* data structure declarations */
/*                             */
// used to pass message through the workqueue
struct msg_work
{
  uint8_t *msg;
  int len;
  struct work_struct work; // work struct used by the workqueue
};

/*                  */
/* global variables */
/*                  */
// socket
static struct socket *udpsocket = NULL;
static struct msghdr send_msg_struct;

// thread for receiving
static DECLARE_COMPLETION(threadcomplete);
static struct task_struct *thread_recv = NULL;

// workqueue for sending
static struct workqueue_struct *wq_send = NULL;

// module parameters
static char peer_addr[16] = "127.0.0.1";
module_param_string(peer_addr, peer_addr, sizeof(peer_addr), 0660);
MODULE_PARM_DESC(peer_addr, "IP Address of the peer");

static __be16 peer_port = 1984;
module_param(peer_port, ushort, 0660);
MODULE_PARM_DESC(peer_port, "UDP Port of the peer");

/*                          */
/* function implementations */
/*                          */
static int com_thread(void *data)
{
  struct msghdr msg;
  struct kvec iov;
  int len;
  #if DEBUG_TIME
  struct timeval t;
  #endif

  allow_signal(SIGTERM);

  // prepare struct for receiving
  msg.msg_name = NULL;
  msg.msg_namelen = 0;
  msg.msg_control = NULL;
  msg.msg_controllen = 0;

  // receive loop
  while(!signal_pending(current)) {
    // get some memory (GFP_KERNEL is ok here because we are in process context)
    uint8_t *buf = kmalloc(WIFI_EMU_MSG_LEN, GFP_KERNEL);
    if(!buf) {
      printk(KERN_ERR "wifi-emu-udp: kmalloc failed\n");
      return -ENOMEM;
    }
    iov.iov_base = buf; // !: iov_base is incremented by recvmsg, thus buf has to be used to get the received data
    iov.iov_len = WIFI_EMU_MSG_LEN;

    // try to receive data
    len = kernel_recvmsg(udpsocket, &msg, &iov, 1, WIFI_EMU_MSG_LEN, 0);

    #if DEBUG_TIME
    if(((struct wifi_emu_msg_header *)buf)->type == WIFI_EMU_MSG_FRAME) {
      do_gettimeofday(&t);
      printk(KERN_INFO "KERN_GotFromSim\t%lu\t%06lu\n", t.tv_sec, t.tv_usec);
    }
    #endif

    if(len >= 0) {
      // pass data to wifi-emu module
      // (some framing mechanism in order to prevent passing chunks of data would be better,
      //   but is left out here in order to keep it simple)
      wifi_emu_recv_msg(buf, len);
    }
    else {
      // throw away data
      kfree(buf);
      if(!signal_pending(current))
        printk(KERN_WARNING "wifi-emu-udp: kernel_recvmsg failed (errno %i)\n", len);
    }
    
  }

  complete(&threadcomplete);
  return 0;
}

static void wifi_emu_send_msg(uint8_t *msg, int len)
{
  // don't send message directly, but put it in a workqueue instead
  // (kernel_sendmsg is not allowed from interrupt context)
  
  // allocate struct we're going to enqueue
  struct msg_work *workm = kmalloc(sizeof(struct msg_work), GFP_ATOMIC);
  if(!workm) {
    printk(KERN_ERR "wifi-emu-udp: kmalloc failed\n");
    return;
  }

  // copy message pointer (send_msg_wq will free it)
  workm->msg = msg;
  workm->len = len;

  // initalize work struct
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20))
  INIT_WORK(&workm->work, send_msg_wq);
#else
  INIT_WORK(&workm->work, send_msg_wq, workm);
#endif

  // enqueue it
  queue_work(wq_send, &workm->work);
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20))
static void send_msg_wq(struct work_struct *w)
{
  struct msg_work *workm = container_of(w, struct msg_work, work);
#else
static void send_msg_wq(void *data)
{
  struct msg_work *workm = (struct msg_work *)data;
#endif
  #if DEBUG_TIME
  struct timeval t;
  #endif

  // now really send the message
  struct kvec iov;
  iov.iov_base = workm->msg;
  iov.iov_len = workm->len;

  #if DEBUG_TIME
  if(((struct wifi_emu_msg_header *)(workm->msg))->type == WIFI_EMU_MSG_FRAME) {
    do_gettimeofday(&t);
    printk(KERN_INFO "KERN_SendToSim\t%lu\t%06lu\n", t.tv_sec, t.tv_usec);
  }
  #endif

  if(kernel_sendmsg(udpsocket, &send_msg_struct, &iov, 1, workm->len) != workm->len)
    printk(KERN_WARNING "wifi-emu-udp: kernel_sendmsg failed\n");

  // free struct and message
  kfree(workm->msg);
  kfree(workm);
}

static int __init wifi_emu_udp_init(void)
{
  struct sockaddr_in addr;
  int err;

  printk(KERN_INFO "wifi-emu-udp: Loading Wifi Emulation Driver (UDP part)\n");

  // create socket
  err = sock_create(PF_INET, SOCK_DGRAM, IPPROTO_UDP, &udpsocket);
  if(err < 0) {
    goto error;
  }


  // bind socket
  // We use a connected socket, in which the kernel chooses our port.
  // This means that we have to send the first message, so that our peer knows us.
  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = INADDR_ANY;
  addr.sin_port = htons(0);
  err = udpsocket->ops->bind(udpsocket, (struct sockaddr *)&addr, sizeof(addr));
  if(err) {
    err = -EIO;
    goto error;
  }

  // set peer address
  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = in_aton(peer_addr);
  addr.sin_port = htons(peer_port);
  err = udpsocket->ops->connect(udpsocket, (struct sockaddr *)&addr, sizeof(addr), 0);

  // initialize struct used for sending
  send_msg_struct.msg_name = NULL; //connected socket  
  send_msg_struct.msg_namelen = 0;
  send_msg_struct.msg_control = NULL;
  send_msg_struct.msg_controllen = 0;

  // create workqueue (sending)
  wq_send = create_singlethread_workqueue("wifi-emu-udp-send");
  if(!wq_send) {
    err = -EIO;
    goto error;
  }

  // create thread (receiving)
  thread_recv = kthread_run(com_thread, NULL, "wifi-emu-udp-recv");
  if(IS_ERR(thread_recv)) {
    err = -EIO;
    goto error;
  }

  // initialize wifi-emu module
  err = wifi_emu_init(&wifi_emu_send_msg);
  if(err < 0) {
    goto error;
  }

  return 0;

error:
  free_resources();
  return err;
}


static void __exit wifi_emu_udp_cleanup(void)
{
  printk(KERN_INFO "wifi-emu-udp: Unloading Wifi Emulation Driver (UDP Part)\n");

  // clean up wifi-emu module
  wifi_emu_cleanup();

  // free all resources
  free_resources();
}

static inline void free_resources(void)
{
  // stop thread
  if(thread_recv) {
    send_sig(SIGTERM, thread_recv, 0);
    wait_for_completion(&threadcomplete);
  }

  // destroy workqueue
  if(wq_send)
    destroy_workqueue(wq_send);

  // remove socket (as last, because code before uses socket)
  if(udpsocket)
    sock_release(udpsocket);
}


module_init(wifi_emu_udp_init);
module_exit(wifi_emu_udp_cleanup);


