/*
 * Copyright (c) 2010 RWTH Aachen University
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Hendrik vom Lehn <vomlehn@cs.rwth-aachen.de>
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/wireless.h>
#include <linux/wait.h>
#include <linux/sched.h>
#include <linux/if_arp.h>
#include <linux/ethtool.h>
#include <linux/time.h>
#include <net/iw_handler.h>

#include "wifi-emu.h"
#include "wifi-emu-msg.h"

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Hendrik vom Lehn <vomlehn@cs.rwth-aachen.de>");
MODULE_DESCRIPTION("Wifi Emulation Driver (generic part)");


/*         */
/* defines */
/*         */
#define DEBUG_TIME 0
#define WIFI_EMU_RESP_TIMEOUT HZ*2 // 2 seconds
#define BUFFER_MIN 2500 // minimum free space in buffer before queue is stopped
#define MIN(a,b) ((a)>(b)?(b):(a))
#define MIN_MTU 68 // same as eth_change_mtu (eth.c)
#define DEFAULT_MTU 1500 // will be set upon initialization

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27))
#define IWE_HELPER(A,B,C,D,E) A,B,C,D,E
#else
#define IWE_HELPER(A,B,C,D,E) B,C,D,E
#endif

/*                                 */
/* function prototype declarations */
/*                                 */
// wireless handler-functions
static int wifi_emu_get_name(struct net_device *dev, struct iw_request_info *info, char *name, char *extra);
static int wifi_emu_set_freq(struct net_device *dev, struct iw_request_info *info, struct iw_freq *freq, char *extra);
static int wifi_emu_get_freq(struct net_device *dev, struct iw_request_info *info, struct iw_freq *freq, char *extra);
static int wifi_emu_set_mode(struct net_device *dev, struct iw_request_info *info, __u32 *mode, char *extra);
static int wifi_emu_get_mode(struct net_device *dev, struct iw_request_info *info, __u32 *mode, char *extra);
static int wifi_emu_get_range(struct net_device *dev, struct iw_request_info *info, struct iw_point *wrq, char *extra);
static int wifi_emu_set_wap(struct net_device *dev, struct iw_request_info *info, struct sockaddr *ap_addr, char *extra);
static int wifi_emu_get_wap(struct net_device *dev, struct iw_request_info *info, struct sockaddr *ap_addr, char *extra);
static int wifi_emu_set_scan(struct net_device *dev, struct iw_request_info *info, struct iw_point *srq, char *extra);
static int wifi_emu_get_scan(struct net_device *dev, struct iw_request_info *info, struct iw_point *srq, char *extra);
static int wifi_emu_set_essid(struct net_device *dev, struct iw_request_info *info, struct iw_point *data, char *essid);
static int wifi_emu_get_essid(struct net_device *dev, struct iw_request_info *info, struct iw_point *data, char *essid);
static int wifi_emu_get_nick(struct net_device *dev, struct iw_request_info *info, struct iw_point *data, char *nick);
static int wifi_emu_set_rate(struct net_device *dev, struct iw_request_info *info, struct iw_param *rrq, char *extra);
static int wifi_emu_get_rate(struct net_device *dev, struct iw_request_info *info, struct iw_param *rrq, char *extra);
static int wifi_emu_set_spy(struct net_device *dev, struct iw_request_info *info, struct iw_point *data, char *extra);
static int wifi_emu_get_fer(struct net_device *dev, struct iw_request_info *info, union iwreq_data *wrqu, char *extra);


// return wireless statistics
static struct iw_statistics *wifi_emu_get_wireless_stats(struct net_device *dev);

// "ifconfig up"
static int wifi_emu_net_open(struct net_device *dev);
// "ifconfig down"
static int wifi_emu_net_stop(struct net_device *dev);
// "ifconfig"
static struct net_device_stats *wifi_emu_get_stats(struct net_device *dev);
// outgoing packets
static int wifi_emu_hard_start_xmit(struct sk_buff *skb, struct net_device *dev);
// change mtu
static int wifi_emu_change_mtu(struct net_device *dev, int new_mtu);

// called by wifi-emu-X
int wifi_emu_init(void (*f_send_msg)(uint8_t *msg, int len));
void wifi_emu_cleanup(void);

// send function (implemented in wifi-emu-X)
static void (*wifi_emu_send_msg)(uint8_t *msg, int len); // function pointer, not implemented here
// helper function which sends message and waits for setter-response
static inline int wifi_emu_send_and_wait(uint8_t *msg, int len);

// receives message and dispatches by type
void wifi_emu_recv_msg(uint8_t *msg, int len);
// receive message handler functions
static void wifi_emu_recv_resp_register(struct wifi_emu_msg_resp_register *msg);
static void wifi_emu_recv_stats(struct wifi_emu_msg_stats *msg);
static void wifi_emu_recv_resp_setter(struct wifi_emu_msg_resp_setter *msg);
static void wifi_emu_recv_frame_sent(struct wifi_emu_msg_frame_sent *msg);
static void wifi_emu_recv_frame(struct wifi_emu_msg_frame *msg, int len);
static void wifi_emu_recv_resp_scan(struct wifi_emu_msg_resp_scan *msg, int len);
static void wifi_emu_recv_resp_fer(struct wifi_emu_msg_resp_fer *msg);
static void wifi_emu_recv_spy_update(struct wifi_emu_msg_spy_update *msg);

// performs updates based on received statistics data
static void do_update_stats(void);

/*                             */
/* data structure declarations */
/*                             */
// holds device specific information (is allocated through alloc_netdev)
struct wifi_emu_priv
{
  struct net_device *dev;
  struct net_device_stats stats;
  struct wifi_emu_stats wstats; // wireless stats from wifi-emu-msg.h
  struct mutex wstats_mutex;
  struct iw_public_data wireless_data;
  struct iw_spy_data spy_data;
  long long buffer_free; // virtual buffer which stores how many bytes may be queued in the simulation (for sent frames)
  int device_up; 
  int max_mtu; // stores the maximum MTU size sent by the simulation
};

/*                  */
/* global variables */
/*                  */
static struct wifi_emu_priv *wifi_emu_mypriv = NULL; // real drivers store this in hardware, but we don't have any :(

// module parameters
static uint16_t client_id = 1;
module_param(client_id, ushort, 0660);
MODULE_PARM_DESC(client_id, "ID used for registration at the network simulation");

static unsigned long buffer_size = 65536;
module_param(buffer_size, ulong, 0660);
MODULE_PARM_DESC(buffer_size, "size of the virtual send buffer");

static int enable_buffer = 1;
module_param(enable_buffer, bool, 0660);
MODULE_PARM_DESC(enable_buffer, "enables the virtual send buffer");

// helper variables used throughout the code (device unspecific)
static struct iw_statistics wifi_emu_iwstats; // only used as return value, updated only when needed (no mutex necessary)
static wait_queue_head_t wifi_emu_wq_resp; // waitqueue used while waiting for responses (registration and setter)
static struct mutex wifi_emu_wq_resp_mutex;
static struct wifi_emu_msg_resp_register *wifi_emu_reg = NULL; // holds response after registering
static struct mutex wifi_emu_reg_mutex;
static struct wifi_emu_msg_resp_setter *wifi_emu_setter = NULL; // holds response of set messages
static struct mutex wifi_emu_setter_mutex;
static struct wifi_emu_msg_resp_scan *wifi_emu_scan = NULL; // holds response of a scan request
static struct mutex wifi_emu_scan_mutex;
static struct wifi_emu_msg_resp_fer *wifi_emu_fer = NULL; // holds response of a FER request
static struct mutex wifi_emu_fer_mutex;

// wireless handlers
// (wireless extensions are explained in include/net/iw_handler.h and include/linux/wireless.h)
static const iw_handler wifi_emu_iw_handler[] = {
  [IW_IOCTL_IDX(SIOCGIWNAME)] = (iw_handler) wifi_emu_get_name,
  [IW_IOCTL_IDX(SIOCSIWFREQ)] = (iw_handler) wifi_emu_set_freq,
  [IW_IOCTL_IDX(SIOCGIWFREQ)] = (iw_handler) wifi_emu_get_freq,
  [IW_IOCTL_IDX(SIOCSIWMODE)] = (iw_handler) wifi_emu_set_mode,
  [IW_IOCTL_IDX(SIOCGIWMODE)] = (iw_handler) wifi_emu_get_mode,
  [IW_IOCTL_IDX(SIOCGIWRANGE)] = (iw_handler) wifi_emu_get_range,
  [IW_IOCTL_IDX(SIOCSIWSPY)] = (iw_handler) wifi_emu_set_spy, // we use the spy functions provided by wext.c but intercept the set_spy in order to tell the simulation which addresses we want
  [IW_IOCTL_IDX(SIOCGIWSPY)] = (iw_handler) iw_handler_get_spy,
  [IW_IOCTL_IDX(SIOCSIWTHRSPY)] = (iw_handler) iw_handler_set_thrspy,
  [IW_IOCTL_IDX(SIOCGIWTHRSPY)] = (iw_handler) iw_handler_get_thrspy,
  [IW_IOCTL_IDX(SIOCSIWAP)] = (iw_handler) wifi_emu_set_wap,
  [IW_IOCTL_IDX(SIOCGIWAP)] = (iw_handler) wifi_emu_get_wap,
  [IW_IOCTL_IDX(SIOCSIWSCAN)] = (iw_handler) wifi_emu_set_scan,
  [IW_IOCTL_IDX(SIOCGIWSCAN)] = (iw_handler) wifi_emu_get_scan,
  [IW_IOCTL_IDX(SIOCSIWESSID)] = (iw_handler) wifi_emu_set_essid,
  [IW_IOCTL_IDX(SIOCGIWESSID)] = (iw_handler) wifi_emu_get_essid,
  [IW_IOCTL_IDX(SIOCGIWNICKN)] = (iw_handler) wifi_emu_get_nick,
  [IW_IOCTL_IDX(SIOCSIWRATE)] = (iw_handler) wifi_emu_set_rate,
  [IW_IOCTL_IDX(SIOCGIWRATE)] = (iw_handler) wifi_emu_get_rate,

};

// private wireless handlers
// can be accessed through the iwpriv command of the wireless-tools
// (iwpriv.c contains some good documentation on how the private handlers work)
static const iw_handler wifi_emu_iw_handler_priv[] = {
  [1] = (iw_handler) wifi_emu_get_fer,
};
static const struct iw_priv_args wifi_emu_iw_handler_priv_args[] = {
  // cmd, set_args, get_args, name
  {SIOCIWFIRSTPRIV+1, IW_PRIV_TYPE_NONE, IW_PRIV_TYPE_BYTE | IW_PRIV_SIZE_FIXED | 1, "get_fer"},
};
static const struct iw_handler_def wifi_emu_iw_handlers = {
  .num_standard           = ARRAY_SIZE(wifi_emu_iw_handler),
  .num_private            = ARRAY_SIZE(wifi_emu_iw_handler_priv),
  .num_private_args       = ARRAY_SIZE(wifi_emu_iw_handler_priv_args),
  .standard               = wifi_emu_iw_handler,
  .private                = wifi_emu_iw_handler_priv,
  .private_args           = wifi_emu_iw_handler_priv_args,
  .get_wireless_stats     = wifi_emu_get_wireless_stats,
};

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,29))
static const struct net_device_ops wifi_emu_netdev_ops = {
  .ndo_open  = wifi_emu_net_open,
  .ndo_stop  = wifi_emu_net_stop,
  .ndo_get_stats  = wifi_emu_get_stats,
  .ndo_start_xmit  = wifi_emu_hard_start_xmit,
  .ndo_change_mtu = wifi_emu_change_mtu,
};
#endif

/*                          */
/* function implementations */
/*                          */

static int wifi_emu_get_name(struct net_device *dev, struct iw_request_info *info, char *name, char *extra)
{
  struct wifi_emu_priv *priv = (struct wifi_emu_priv *)netdev_priv(dev);

  switch(priv->wstats.standard) {
    case WIFI_EMU_STANDARD_UNASSOCIATED:
      strcpy(name, "unassociated"); return 0;
    case WIFI_EMU_STANDARD_80211A:
      strcpy(name, "IEEE 802.11a"); return 0;
    case WIFI_EMU_STANDARD_80211B:
      strcpy(name, "IEEE 802.11b"); return 0;
  }

  strcpy(name, "");
  return 0;
}

static int wifi_emu_set_freq(struct net_device *dev, struct iw_request_info *info, struct iw_freq *freq, char *extra)
{
  struct wifi_emu_msg_req_setchannel *msg;

  if(freq->e != 0) {
    printk(KERN_WARNING "wifi-emu: Frequency setting only supported by channel\n");
    return -EINVAL;
  }

  msg = kmalloc(sizeof(struct wifi_emu_msg_req_setchannel), GFP_ATOMIC);
  if(!msg) {
    printk(KERN_ERR "wifi-emu: kmalloc failed\n");
    return -ENOMEM;
  }

  msg->h.type = WIFI_EMU_MSG_REQ_SETCHANNEL;
  msg->h.client_id = htons(client_id);
  msg->channel = htons(freq->m);

  return wifi_emu_send_and_wait((uint8_t *)msg, sizeof(*msg));
}

static int wifi_emu_get_freq(struct net_device *dev, struct iw_request_info *info, struct iw_freq *freq, char *extra)
{
  struct wifi_emu_priv *priv = (struct wifi_emu_priv *)netdev_priv(dev);

  mutex_lock(&priv->wstats_mutex);
  switch(priv->wstats.standard) {
    case WIFI_EMU_STANDARD_80211A:
      freq->e = 1;
      freq->m = (5000 + 5 * (ntohs(priv->wstats.channel) - 1)) * 100;
      break;
    case WIFI_EMU_STANDARD_80211B:
      freq->e = 1;
      freq->m = (2412 + 5 * (ntohs(priv->wstats.channel) - 1)) * 100;
      break;
    default:
      freq->m = 0;
      freq->e = 0;
      break;
  }

  mutex_unlock(&priv->wstats_mutex);
  return 0;
}

static int wifi_emu_set_mode(struct net_device *dev, struct iw_request_info *info, __u32 *mode, char *extra)
{
  struct wifi_emu_msg_req_setmode *msg;

  msg = kmalloc(sizeof(struct wifi_emu_msg_req_setmode), GFP_ATOMIC);
  if(!msg) {
    printk(KERN_ERR "wifi-emu: kmalloc failed\n");
    return -ENOMEM;
  }

  msg->h.type = WIFI_EMU_MSG_REQ_SETMODE;
  msg->h.client_id = htons(client_id);
  
  switch(*mode) {
    case IW_MODE_MASTER:
      msg->mode = WIFI_EMU_MODE_MASTER; break;
    case IW_MODE_ADHOC:
      msg->mode = WIFI_EMU_MODE_ADHOC; break;
    case IW_MODE_MONITOR:
      msg->mode = WIFI_EMU_MODE_MONITOR; break;
    default:
      printk(KERN_WARNING "wifi-emu: Trying to set unsupported mode\n");
      kfree(msg);
      return -EINVAL;
  }

  return wifi_emu_send_and_wait((uint8_t *)msg, sizeof(*msg));
}

static int wifi_emu_get_mode(struct net_device *dev, struct iw_request_info *info, __u32 *mode, char *extra)
{
  struct wifi_emu_priv *priv = (struct wifi_emu_priv *)netdev_priv(dev);

  switch(priv->wstats.mode) {
    case WIFI_EMU_MODE_MASTER:
      *mode = IW_MODE_MASTER; break;
    case WIFI_EMU_MODE_ADHOC:
      *mode = IW_MODE_ADHOC; break;
    case WIFI_EMU_MODE_MONITOR:
      *mode = IW_MODE_MONITOR; break;
    default:  
      *mode = IW_MODE_AUTO; break;
  }

  return 0;
}

static int wifi_emu_get_range(struct net_device *dev, struct iw_request_info *info, struct iw_point *wrq, char *extra)
{
  struct iw_range *range = (struct iw_range *)extra;

  wrq->length = sizeof(struct iw_range);
  memset(range, 0, sizeof(struct iw_range));
  range->we_version_compiled = WIRELESS_EXT;
  range->we_version_source = 20;

  range->max_qual.qual = 100; // 0..100
  range->max_qual.level = 0; // dBm
  range->max_qual.noise = 0; // dBm
  range->max_qual.updated = IW_QUAL_ALL_UPDATED;

  range->event_capa[0] = IW_EVENT_CAPA_K_0
    | IW_EVENT_CAPA_MASK(SIOCGIWTHRSPY)
    | IW_EVENT_CAPA_MASK(SIOCGIWAP)
    | IW_EVENT_CAPA_MASK(SIOCGIWSCAN);
  range->event_capa[1] = IW_EVENT_CAPA_K_1;

  return 0;
}

static int wifi_emu_set_wap(struct net_device *dev, struct iw_request_info *info, struct sockaddr *ap_addr, char *extra)
{
  struct wifi_emu_msg_req_setwap *msg;

        if (ap_addr->sa_family != ARPHRD_ETHER)
    return -EINVAL;

  msg = kmalloc(sizeof(struct wifi_emu_msg_req_setwap), GFP_ATOMIC);
  if(!msg) {
    printk(KERN_ERR "wifi-emu: kmalloc failed\n");
    return -ENOMEM;
  }

  msg->h.type = WIFI_EMU_MSG_REQ_SETWAP;
  msg->h.client_id = htons(client_id);
  memcpy(msg->remote_mac, ap_addr->sa_data, WIFI_EMU_MAC_LEN);

  return wifi_emu_send_and_wait((uint8_t *)msg, sizeof(*msg));
}

static int wifi_emu_get_wap(struct net_device *dev, struct iw_request_info *info, struct sockaddr *ap_addr, char *extra)
{
  struct wifi_emu_priv *priv = (struct wifi_emu_priv *)netdev_priv(dev);
  
  mutex_lock(&priv->wstats_mutex);
  if(priv->wstats.standard !=  WIFI_EMU_STANDARD_UNASSOCIATED) {
    ap_addr->sa_family = ARPHRD_ETHER;
    memcpy(ap_addr->sa_data,priv->wstats.remote_mac, WIFI_EMU_MAC_LEN);
  }
  else {
    memset(ap_addr->sa_data, 0, WIFI_EMU_MAC_LEN);
  }
  mutex_unlock(&priv->wstats_mutex);

  return 0;
}

static int wifi_emu_set_scan(struct net_device *dev, struct iw_request_info *info, struct iw_point *srq, char *extra)
{
  struct iw_scan_req *req;
  struct wifi_emu_msg_req_scan *msg;

  msg = kmalloc(sizeof(struct wifi_emu_msg_req_scan), GFP_ATOMIC);
  if(!msg) {
    printk(KERN_ERR "wifi-emu: kmalloc failed\n");
    return -ENOMEM;
  }

  msg->h.type = WIFI_EMU_MSG_REQ_SCAN;
  msg->h.client_id = htons(client_id);

  msg->active = 1;
  msg->essid_len = 0;

  if(srq->length == sizeof(struct iw_scan_req)) {
    req = (struct iw_scan_req *)extra;

    if(srq->flags & IW_SCAN_THIS_ESSID) {
      msg->essid_len = MIN(req->essid_len, WIFI_EMU_ESSID_LEN);
      memcpy(msg->essid, req->essid, msg->essid_len);
    }

    if(req->scan_type == IW_SCAN_TYPE_PASSIVE) {
      msg->active = 0;
    }
  }

  mutex_lock(&wifi_emu_scan_mutex);
  wifi_emu_scan = NULL;
  mutex_unlock(&wifi_emu_scan_mutex);

  wifi_emu_send_msg((uint8_t *)msg, sizeof(*msg));

  return 0;
}

static int wifi_emu_get_scan(struct net_device *dev, struct iw_request_info *info, struct iw_point *srq, char *extra)
{
  uint8_t i;
  struct iw_event iwe;
  struct wifi_emu_scan_entry *entry;
  char *current_ev = extra;

  mutex_lock(&wifi_emu_scan_mutex);
  if(wifi_emu_scan == NULL) {
    mutex_unlock(&wifi_emu_scan_mutex);
    return -EAGAIN;
  }

  for(i=0; i<wifi_emu_scan->num_entries; i++) {
    entry = &wifi_emu_scan->data[i];

    // MAC address
    iwe.cmd= SIOCGIWAP;
    iwe.u.ap_addr.sa_family = ARPHRD_ETHER;
    memcpy(iwe.u.ap_addr.sa_data, entry->remote_mac, WIFI_EMU_MAC_LEN);
    current_ev = iwe_stream_add_event(IWE_HELPER(info, current_ev, extra + IW_SCAN_MAX_DATA, &iwe, IW_EV_ADDR_LEN));

    // ESSID
    iwe.cmd = SIOCGIWESSID;
    iwe.u.data.length = MIN(entry->essid_len, WIFI_EMU_ESSID_LEN);
    iwe.u.data.flags = 1;
    current_ev = iwe_stream_add_point(IWE_HELPER(info, current_ev, extra + IW_SCAN_MAX_DATA, &iwe, entry->essid));

    // mode
    iwe.cmd = SIOCGIWMODE;
    switch(entry->mode) {
      case WIFI_EMU_MODE_MASTER:
        iwe.u.mode = IW_MODE_MASTER; break;
      case WIFI_EMU_MODE_ADHOC:
        iwe.u.mode = IW_MODE_ADHOC; break;
      case WIFI_EMU_MODE_MONITOR:
        iwe.u.mode = IW_MODE_MONITOR; break;
      default:
        iwe.u.mode = IW_MODE_AUTO; break;
    }
    current_ev = iwe_stream_add_event(IWE_HELPER(info, current_ev, extra + IW_SCAN_MAX_DATA, &iwe, IW_EV_UINT_LEN));

    // channel
    iwe.cmd = SIOCGIWFREQ;
    iwe.u.freq.m = ntohs(entry->channel);
    iwe.u.freq.e = 0;
    current_ev = iwe_stream_add_event(IWE_HELPER(info, current_ev, extra + IW_SCAN_MAX_DATA, &iwe, IW_EV_FREQ_LEN));

    // quality
    iwe.cmd = IWEVQUAL;
    iwe.u.qual.qual = entry->qual.qual;
    iwe.u.qual.level = entry->qual.level;
    iwe.u.qual.noise = entry->qual.noise;
    current_ev = iwe_stream_add_event(IWE_HELPER(info, current_ev, extra + IW_SCAN_MAX_DATA, &iwe, IW_EV_QUAL_LEN));
  }

  srq->length = (current_ev - extra);
  srq->flags = 0;

  // free scan message
  kfree(wifi_emu_scan);
  wifi_emu_scan = NULL;
  mutex_unlock(&wifi_emu_scan_mutex);
  
  return 0;
}

static int wifi_emu_set_essid(struct net_device *dev, struct iw_request_info *info, struct iw_point *data, char *essid)
{
  struct wifi_emu_msg_req_setessid *msg;

  if(data->length > WIFI_EMU_ESSID_LEN)
    return -EINVAL;

  msg = kmalloc(sizeof(struct wifi_emu_msg_req_setessid), GFP_ATOMIC);
  if(!msg) {
    printk(KERN_ERR "wifi-emu: kmalloc failed\n");
    return -ENOMEM;
  }

  msg->h.type = WIFI_EMU_MSG_REQ_SETESSID;
  msg->h.client_id = htons(client_id);
  msg->essid_len = MIN(data->length, WIFI_EMU_ESSID_LEN);
  memcpy(msg->essid, essid, msg->essid_len);

  return wifi_emu_send_and_wait((uint8_t *)msg, sizeof(*msg));
}

static int wifi_emu_get_essid(struct net_device *dev, struct iw_request_info *info, struct iw_point *data, char *essid)
{
  struct wifi_emu_priv *priv = (struct wifi_emu_priv *)netdev_priv(dev);

  mutex_lock(&priv->wstats_mutex);
  if(priv->wstats.standard !=  WIFI_EMU_STANDARD_UNASSOCIATED) {
    data->length = MIN(priv->wstats.essid_len, WIFI_EMU_ESSID_LEN);
    memcpy(essid, priv->wstats.essid, data->length);
    data->flags = 1;
  }
  else {
    data->length = 0;
    data->flags = 0;
  }
  mutex_unlock(&priv->wstats_mutex);

  return 0;
}

static int wifi_emu_get_nick(struct net_device *dev, struct iw_request_info *info, struct iw_point *data, char *nick)
{
  strcpy(nick, "wifi-emu");
  data->flags = 1;
  data->length = strlen(nick);
  return 0;
}

static int wifi_emu_set_rate(struct net_device *dev, struct iw_request_info *info, struct iw_param *rrq, char *extra)
{
  struct wifi_emu_msg_req_setrate *msg;

  msg = kmalloc(sizeof(struct wifi_emu_msg_req_setrate), GFP_ATOMIC);
  if(!msg) {
    printk(KERN_ERR "wifi-emu: kmalloc failed\n");
    return -ENOMEM;
  }

  msg->h.type = WIFI_EMU_MSG_REQ_SETRATE;
  msg->h.client_id = htons(client_id);
  msg->bitrate = rrq->value;

  return wifi_emu_send_and_wait((uint8_t *)msg, sizeof(*msg));
}

static int wifi_emu_get_rate(struct net_device *dev, struct iw_request_info *info, struct iw_param *rrq, char *extra)
{
  struct wifi_emu_priv *priv = (struct wifi_emu_priv *)netdev_priv(dev);

  mutex_lock(&priv->wstats_mutex);
  rrq->fixed = 0;
  rrq->disabled = 0;
  if(priv->wstats.standard !=  WIFI_EMU_STANDARD_UNASSOCIATED)
    rrq->value = ntohl(priv->wstats.bitrate);
  else
    rrq->value = 0;
  mutex_unlock(&priv->wstats_mutex);

  return 0;
}

static int wifi_emu_set_spy(struct net_device *dev, struct iw_request_info *info, struct iw_point *data, char *extra)
{
  struct wifi_emu_msg_req_setspy *msg;
  int res_sim;
  uint8_t i;
  int msg_len = sizeof(struct wifi_emu_msg_req_setspy)+WIFI_EMU_MAC_LEN*data->length;

  msg = kmalloc(msg_len, GFP_ATOMIC);
  if(!msg) {
    printk(KERN_ERR "wifi-emu: kmalloc failed\n");
    return -ENOMEM;
  }

  msg->h.type = WIFI_EMU_MSG_REQ_SETSPY;
  msg->h.client_id = htons(client_id);

  msg->num_addr = data->length;
  for(i=0; i<msg->num_addr; i++)
    memcpy(msg->addr[i], ((struct sockaddr *)extra)[i].sa_data, WIFI_EMU_MAC_LEN);

  res_sim = wifi_emu_send_and_wait((uint8_t *)msg, msg_len);
  if(res_sim != 0)
    return res_sim;

  return iw_handler_set_spy(dev, info, (union iwreq_data *)data, extra);
}

static int wifi_emu_get_fer(struct net_device *dev, struct iw_request_info *info, union iwreq_data *wrqu, char *extra)
{
  uint8_t fer;
  struct wifi_emu_msg_req_fer *msg_req;

  // create request
  msg_req = kmalloc(sizeof(struct wifi_emu_msg_req_fer), GFP_ATOMIC);
  if(!msg_req) {
    printk(KERN_ERR "wifi-emu: kmalloc failed\n");
    return -ENOMEM;
  }

  msg_req->h.type = WIFI_EMU_MSG_REQ_FER;
  msg_req->h.client_id = htons(client_id);

  // send message
  wifi_emu_send_msg((uint8_t *)msg_req, sizeof(*msg_req));

  // wait for response
  mutex_lock(&wifi_emu_wq_resp_mutex);
  init_waitqueue_head(&wifi_emu_wq_resp);
  if(!wait_event_timeout(wifi_emu_wq_resp, (wifi_emu_fer != NULL), WIFI_EMU_RESP_TIMEOUT)) {
    printk(KERN_ERR "wifi-emu: Timeout while waiting for FER\n");
    mutex_unlock(&wifi_emu_wq_resp_mutex);
    return -EIO;
  }
  mutex_unlock(&wifi_emu_wq_resp_mutex);

  // check response
  mutex_lock(&wifi_emu_fer_mutex);
  if(wifi_emu_fer == NULL) {
    printk(KERN_ERR "wifi-emu: wifi_emu_fer changed to NULL (should not happen)\n");
    mutex_unlock(&wifi_emu_fer_mutex);
    return -EFAULT;
  }

  // get FER from response
  fer = wifi_emu_fer->fer;

  // free message
  kfree(wifi_emu_fer);
  wifi_emu_fer = NULL;
  mutex_unlock(&wifi_emu_fer_mutex);

  // give FER to caller of ioctl
  *extra = fer;

  return 0;
}

static struct iw_statistics *wifi_emu_get_wireless_stats(struct net_device *dev)
{
  struct wifi_emu_priv *priv = (struct wifi_emu_priv*)netdev_priv(dev);

  memset(&wifi_emu_iwstats, 0, sizeof(struct iw_statistics));

  mutex_lock(&priv->wstats_mutex);
  if(priv->wstats.standard !=  WIFI_EMU_STANDARD_UNASSOCIATED) {
    wifi_emu_iwstats.qual.qual = priv->wstats.qual.qual;
    wifi_emu_iwstats.qual.level = priv->wstats.qual.level;
    wifi_emu_iwstats.qual.noise = priv->wstats.qual.noise;
    wifi_emu_iwstats.qual.updated = IW_QUAL_ALL_UPDATED | IW_QUAL_DBM; // level and noise are dBm
  }
  else {
    wifi_emu_iwstats.qual.qual = 0;
    wifi_emu_iwstats.qual.level = 0;
    wifi_emu_iwstats.qual.noise = 0;
    wifi_emu_iwstats.qual.updated = IW_QUAL_ALL_UPDATED;
  }
  mutex_unlock(&priv->wstats_mutex);

  return &wifi_emu_iwstats;
}

static int wifi_emu_net_open(struct net_device *dev) 
{
  struct wifi_emu_msg_req_setifup *msg;
  int send_val;

  msg = kmalloc(sizeof(struct wifi_emu_msg_req_setifup), GFP_ATOMIC);
  if(!msg) {
    printk(KERN_ERR "wifi-emu: kmalloc failed\n");
    return -ENOMEM;
  }

  msg->h.type = WIFI_EMU_MSG_REQ_SETIFUP;
  msg->h.client_id = htons(client_id);
  msg->up = 1;

  send_val = wifi_emu_send_and_wait((uint8_t *)msg, sizeof(*msg));
  
  if(send_val != 0)
    return send_val;

  netif_start_queue(dev);

  wifi_emu_mypriv->device_up = 1;

  printk(KERN_INFO "wifi-emu: Set device \"up\"\n");
  return 0;
}

static int wifi_emu_net_stop(struct net_device *dev) 
{
  struct wifi_emu_msg_req_setifup *msg;
  int send_val;

  msg = kmalloc(sizeof(struct wifi_emu_msg_req_setifup), GFP_ATOMIC);
  if(!msg) {
    printk(KERN_ERR "wifi-emu: kmalloc failed\n");
    return -ENOMEM;
  }

  msg->h.type = WIFI_EMU_MSG_REQ_SETIFUP;
  msg->h.client_id = htons(client_id);
  msg->up = 0;

  send_val = wifi_emu_send_and_wait((uint8_t *)msg, sizeof(*msg));
  
  if(send_val != 0)
    return send_val;

  netif_stop_queue(dev);

  wifi_emu_mypriv->device_up = 0;

  printk(KERN_INFO "wifi-emu: Set device \"down\"\n");
  return 0;
}

static struct net_device_stats *wifi_emu_get_stats(struct net_device *dev)
{
  struct wifi_emu_priv *wemu = (struct wifi_emu_priv*)netdev_priv(dev);
  return &wemu->stats;
}

static int wifi_emu_hard_start_xmit(struct sk_buff *skb, struct net_device *dev)
{
  struct wifi_emu_priv *priv = (struct wifi_emu_priv*)netdev_priv(dev);
  struct wifi_emu_msg_frame *msg;
  #if DEBUG_TIME
  struct timeval t;
  #endif
  #if DEBUG_TIME
  do_gettimeofday(&t);
  printk(KERN_INFO "KERN_GotFromOS\t%lu\t%06lu\n", t.tv_sec, t.tv_usec);
  #endif

  if(priv->wstats.standard ==  WIFI_EMU_STANDARD_UNASSOCIATED || priv->wstats.mode ==  WIFI_EMU_MODE_MONITOR) {
    wifi_emu_mypriv->stats.tx_dropped++;
    printk(KERN_WARNING "wifi-emu: Cannot send packets while unassociated or in monitor mode\n");
    dev_kfree_skb(skb);
    return NETDEV_TX_OK; // returning anything else indicates that the device is busy and that sending shall be retried
  }

  // check if buffer has enough space 
  if(enable_buffer && wifi_emu_mypriv->buffer_free < skb->len) {

    if(!netif_queue_stopped(wifi_emu_mypriv->dev)) {
      netif_stop_queue(wifi_emu_mypriv->dev);
      printk(KERN_WARNING "wifi-emu: queue was not stopped although buffer is empty!\n");
    }

    printk(KERN_WARNING "wifi-emu: xmit has been called although queue should be stopped!\n");
    return NETDEV_TX_BUSY;
  }

  // send frame
  msg = kmalloc(sizeof(struct wifi_emu_msg_frame)+skb->len, GFP_ATOMIC);
  if(!msg) {
    printk(KERN_ERR "wifi-emu: kmalloc failed\n");
    return -ENOMEM;
  }

  msg->h.type = WIFI_EMU_MSG_FRAME;
  msg->h.client_id = htons(client_id);

  memcpy(msg->data, skb->data, skb->len);
  msg->data_len = htons(skb->len);

  wifi_emu_send_msg((uint8_t *)msg, sizeof(struct wifi_emu_msg_frame)+skb->len);

  // update stats
  wifi_emu_mypriv->stats.tx_packets++;
  wifi_emu_mypriv->stats.tx_bytes += skb->len;

  if(enable_buffer) {
    // update buffer
    wifi_emu_mypriv->buffer_free -= skb->len;

    // stop queue if too few space in buffer to accept next frame
    if(wifi_emu_mypriv->buffer_free < BUFFER_MIN) {
      netif_stop_queue(wifi_emu_mypriv->dev);
    }
  }
  
  // free skb
  dev_kfree_skb(skb);

  return NETDEV_TX_OK;
}

static int wifi_emu_change_mtu(struct net_device *dev, int new_mtu)
{
  if(new_mtu < MIN_MTU || new_mtu > wifi_emu_mypriv->max_mtu)
    return -EINVAL;

  dev->mtu = new_mtu;
  return 0;
}

int wifi_emu_init(void (*f_send_msg)(uint8_t *msg, int len))
{
  int err;
  struct net_device *dev;

  struct wifi_emu_msg_req_register *msg_req_register;

  if(wifi_emu_mypriv != NULL) {
    printk(KERN_ERR "wifi-emu: wifi_emu_mypriv is already initialized\n");
    return -EEXIST; //don't initalize a second time
  }

  printk(KERN_INFO "wifi-emu: Initalizing Wifi Emulation Driver\n");

  mutex_init(&wifi_emu_wq_resp_mutex);
  mutex_init(&wifi_emu_reg_mutex);
  mutex_init(&wifi_emu_setter_mutex);
  mutex_init(&wifi_emu_scan_mutex);
  mutex_init(&wifi_emu_fer_mutex);

  wifi_emu_send_msg = f_send_msg;

  // register at simulation
  msg_req_register = kmalloc(sizeof(struct wifi_emu_msg_req_register), GFP_ATOMIC);
  if(!msg_req_register) {
    printk(KERN_ERR "wifi-emu: kmalloc failed\n");
    return -ENOMEM;
  }

  msg_req_register->h.type = WIFI_EMU_MSG_REQ_REGISTER;
  msg_req_register->h.client_id = htons(client_id);
  if(enable_buffer)
    msg_req_register->virtual_buf = 1;
  else
    msg_req_register->virtual_buf = 0;
  wifi_emu_send_msg((uint8_t *)msg_req_register, sizeof(*msg_req_register));

  // wait for response
  mutex_lock(&wifi_emu_wq_resp_mutex);
  init_waitqueue_head(&wifi_emu_wq_resp);
  if(!wait_event_timeout(wifi_emu_wq_resp, (wifi_emu_reg != NULL), WIFI_EMU_RESP_TIMEOUT)) {
    printk(KERN_ERR "wifi-emu: Timeout while registering at the simulation\n");
    mutex_unlock(&wifi_emu_wq_resp_mutex);
    return -EIO;
  }
  mutex_unlock(&wifi_emu_wq_resp_mutex);

  mutex_lock(&wifi_emu_reg_mutex);
  if(wifi_emu_reg == NULL) {
    printk(KERN_ERR "wifi-emu: wifi_emu_reg changed to NULL (should not happen)\n");
    mutex_unlock(&wifi_emu_reg_mutex);
    return -EFAULT;
  }

  // check response
  if(!wifi_emu_reg->ok) {
    printk(KERN_ERR "wifi-emu: Simulation rejected register message\n");
    kfree(wifi_emu_reg);
    wifi_emu_reg = NULL;
    mutex_unlock(&wifi_emu_reg_mutex);
    return -EACCES;
  }

  // create net device
  dev = alloc_netdev(sizeof(struct wifi_emu_priv), "wemu%d", ether_setup);
  if(!dev) {
    mutex_unlock(&wifi_emu_reg_mutex);
    return -ENOMEM;
  }
  wifi_emu_mypriv = netdev_priv(dev);
  mutex_init(&wifi_emu_mypriv->wstats_mutex);

  wifi_emu_mypriv->buffer_free = buffer_size;
  wifi_emu_mypriv->device_up = 0;
  wifi_emu_mypriv->dev = dev;

  dev->wireless_data = &wifi_emu_mypriv->wireless_data;
  dev->wireless_data->spy_data = &wifi_emu_mypriv->spy_data;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,29))
  dev->netdev_ops = &wifi_emu_netdev_ops;
#else
  dev->open = wifi_emu_net_open;
  dev->stop = wifi_emu_net_stop;
  dev->get_stats = wifi_emu_get_stats;
  dev->hard_start_xmit = wifi_emu_hard_start_xmit;
  dev->change_mtu = wifi_emu_change_mtu;
#endif

  dev->wireless_handlers = &wifi_emu_iw_handlers; // cfg80211 is nicer, but only supported in newer kernels

  // get mac address from register response
  if(WIFI_EMU_MAC_LEN > MAX_ADDR_LEN) {
    printk(KERN_ERR "wifi-emu: Wrong mac address size\n");
    kfree(wifi_emu_reg);
    mutex_unlock(&wifi_emu_reg_mutex);
    return -EIO;
  }
  memcpy(dev->dev_addr, wifi_emu_reg->mac_address, WIFI_EMU_MAC_LEN);

  // set mtu
  wifi_emu_mypriv->max_mtu = ntohs(wifi_emu_reg->mtu);
  dev->mtu = MIN(DEFAULT_MTU, wifi_emu_mypriv->max_mtu);

  // copy statistics included in the response
  mutex_lock(&wifi_emu_mypriv->wstats_mutex);
  memcpy(&wifi_emu_mypriv->wstats, &wifi_emu_reg->wstats, sizeof(struct wifi_emu_stats));
  mutex_unlock(&wifi_emu_mypriv->wstats_mutex);
  do_update_stats();
  
  // free wifi_emu_reg (not needed any more)
  kfree(wifi_emu_reg);
  wifi_emu_reg = NULL;
  mutex_unlock(&wifi_emu_reg_mutex);

  err = register_netdev(dev);
  if(err)
    goto error;

  return 0;

error:
  free_netdev(dev);
  return err;
}


void wifi_emu_cleanup(void)
{
  struct wifi_emu_msg_unregister *msg;

  if(wifi_emu_mypriv == NULL) {
    printk(KERN_ERR "wifi-emu: wifi_emu_mypriv is already NULL (should not happen)\n");
    return;
  }

  printk(KERN_INFO "wifi-emu: Cleaning Wifi Emulation Driver\n");

  // tell kernel to unregister netdevice
  // (brings interface down if needed)
  unregister_netdev(wifi_emu_mypriv->dev);
  free_netdev(wifi_emu_mypriv->dev);

  // tell simulation that we'll stop
  msg = kmalloc(sizeof(struct wifi_emu_msg_unregister), GFP_ATOMIC);
  if(!msg) {
    printk(KERN_ERR "wifi-emu: kmalloc failed\n");
  }
  else {
    msg->h.type = WIFI_EMU_MSG_UNREGISTER;
    msg->h.client_id = htons(client_id);
    wifi_emu_send_msg((uint8_t *)msg, sizeof(*msg));
  }

  // reset global pointers
  wifi_emu_mypriv = NULL;
  wifi_emu_reg = NULL;
  wifi_emu_setter = NULL;
  wifi_emu_scan = NULL;
  wifi_emu_fer = NULL;
}

static inline int wifi_emu_send_and_wait(uint8_t *msg, int len)
{
  // send message
  wifi_emu_send_msg((uint8_t *)msg, len);

  // wait for response
  mutex_lock(&wifi_emu_wq_resp_mutex);
  init_waitqueue_head(&wifi_emu_wq_resp);
  if(!wait_event_timeout(wifi_emu_wq_resp, (wifi_emu_setter != NULL), WIFI_EMU_RESP_TIMEOUT)) {
    printk(KERN_ERR "wifi-emu: Timeout while waiting for response\n");
    mutex_unlock(&wifi_emu_wq_resp_mutex);
    return -EIO;
  }
  mutex_unlock(&wifi_emu_wq_resp_mutex);

  mutex_lock(&wifi_emu_setter_mutex);
  if(wifi_emu_setter == NULL) {
    printk(KERN_ERR "wifi-emu: wifi_emu_setter changed to NULL (should not happen)\n");
    mutex_unlock(&wifi_emu_setter_mutex);
    return -EFAULT;
  }

  // check response
  if(!wifi_emu_setter->ok) {
    printk(KERN_WARNING "wifi-emu: Set request denied by simulation\n");
    kfree(wifi_emu_setter);
    wifi_emu_setter = NULL;
    mutex_unlock(&wifi_emu_setter_mutex);
    return -EINVAL;
  }

  kfree(wifi_emu_setter);
  wifi_emu_setter = NULL;
  mutex_unlock(&wifi_emu_setter_mutex);
  return 0;
}

void wifi_emu_recv_msg(uint8_t *msg, int len)
{
  // check minimum length
  if(len < sizeof(struct wifi_emu_msg_header)) {
    printk(KERN_WARNING "wifi-emu: Received unknown data (%i bytes)\n", len);
    kfree(msg);
    return;
  }

  // check client id
  if(ntohs(((struct wifi_emu_msg_header *)msg)->client_id) != client_id) {
    printk(KERN_WARNING "wifi-emu: Received data for wrong client\n");
    kfree(msg);
    return;
  }

  // branch according to message type
  switch(((struct wifi_emu_msg_header *)msg)->type) {
    case WIFI_EMU_MSG_RESP_REGISTER:
      if(len != sizeof(struct wifi_emu_msg_resp_register)) goto errlength;
      wifi_emu_recv_resp_register((struct wifi_emu_msg_resp_register *)msg);
      return;
    case WIFI_EMU_MSG_STATS:
      if(len != sizeof(struct wifi_emu_msg_stats)) goto errlength;
      wifi_emu_recv_stats((struct wifi_emu_msg_stats *)msg);
      return;
    case WIFI_EMU_MSG_RESP_SETTER:
      if(len != sizeof(struct wifi_emu_msg_resp_setter)) goto errlength;
      wifi_emu_recv_resp_setter((struct wifi_emu_msg_resp_setter *)msg);
      return;
    case WIFI_EMU_MSG_SPY_UPDATE:
      if(len != sizeof(struct wifi_emu_msg_spy_update)) goto errlength;
      wifi_emu_recv_spy_update((struct wifi_emu_msg_spy_update *)msg);
      return;
    case WIFI_EMU_MSG_FRAME_SENT:
      if(len != sizeof(struct wifi_emu_msg_frame_sent)) goto errlength;
      wifi_emu_recv_frame_sent((struct wifi_emu_msg_frame_sent *)msg);
      return;
    case WIFI_EMU_MSG_RESP_FER:
      if(len != sizeof(struct wifi_emu_msg_resp_fer)) goto errlength;
      wifi_emu_recv_resp_fer((struct wifi_emu_msg_resp_fer *)msg);
      return;
    case WIFI_EMU_MSG_FRAME:
      // length is checked by wifi_emu_recv_frame
      wifi_emu_recv_frame((struct wifi_emu_msg_frame *)msg, len);
      return;
    case WIFI_EMU_MSG_RESP_SCAN:
      // length is checked by wifi_emu_recv_resp_scan
      wifi_emu_recv_resp_scan((struct wifi_emu_msg_resp_scan *)msg, len);
      return;
  }

  // We only get here if the switch did not match
  printk(KERN_WARNING "wifi-emu: Received data with wrong type\n");
  kfree(msg);
  return;

errlength:
  printk(KERN_WARNING "wifi-emu: Received data with wrong length\n");
  kfree(msg);
  return;
}

static void wifi_emu_recv_resp_register(struct wifi_emu_msg_resp_register *msg)
{
  // check if not already initalized
  if(wifi_emu_mypriv != NULL) { 
    printk(KERN_ERR "wifi-emu: Received registration message although already initialized");
    kfree(msg);
    return;
  }

  // init method is waiting for wifi_emu_reg to become != NULL
  // (we pass the message received with it)
  mutex_lock(&wifi_emu_reg_mutex);
  wifi_emu_reg = msg;
  mutex_unlock(&wifi_emu_reg_mutex);
  wake_up(&wifi_emu_wq_resp);

  // do not free (wifi_emu_reg points to the message)
  return;
}

static void wifi_emu_recv_stats(struct wifi_emu_msg_stats *msg)
{
  // check if already initalized
  if(wifi_emu_mypriv == NULL) { 
    printk(KERN_ERR "wifi-emu: Received message although not initialized");
    kfree(msg);
    return;
  }

  // copy statistics
  mutex_lock(&wifi_emu_mypriv->wstats_mutex);
  memcpy(&wifi_emu_mypriv->wstats, &msg->wstats, sizeof(struct wifi_emu_stats));
  mutex_unlock(&wifi_emu_mypriv->wstats_mutex);
  do_update_stats();

  kfree(msg);
  return;
}

static void wifi_emu_recv_frame_sent(struct wifi_emu_msg_frame_sent *msg)
{
  // check if already initalized
  if(wifi_emu_mypriv == NULL) { 
    printk(KERN_ERR "wifi-emu: Received message although not initialized");
    kfree(msg);
    return;
  }

  // check if virtual buffer is used at all
  if(!enable_buffer) {
    kfree(msg);
    return;
  }

  // udpate buffer
  wifi_emu_mypriv->buffer_free += ntohs(msg->len);

  // wake queue if buffer has enough space to accept new frames
  if(netif_queue_stopped(wifi_emu_mypriv->dev) && wifi_emu_mypriv->buffer_free >= BUFFER_MIN && wifi_emu_mypriv->device_up) {
    netif_wake_queue(wifi_emu_mypriv->dev);
  }

  kfree(msg);
  return;
}

static void wifi_emu_recv_resp_setter(struct wifi_emu_msg_resp_setter *msg)
{
  // check if already initalized
  if(wifi_emu_mypriv == NULL) { 
    printk(KERN_ERR "wifi-emu: Received message although not initialized");
    kfree(msg);
    return;
  }

  mutex_lock(&wifi_emu_setter_mutex);
  wifi_emu_setter = msg;
  mutex_unlock(&wifi_emu_setter_mutex);

  wake_up(&wifi_emu_wq_resp);

  // do not free (msg is still used)
  return;
}

static void wifi_emu_recv_frame(struct wifi_emu_msg_frame *msg, int len)
{
  struct sk_buff *skb;
  #if DEBUG_TIME
  struct timeval t;
  #endif

  // check if already initalized
  if(wifi_emu_mypriv == NULL) { 
    printk(KERN_ERR "wifi-emu: Received message although not initialized");
    kfree(msg);
    return;
  }

  // check length
  if(ntohs(msg->data_len) > (len - sizeof(struct wifi_emu_msg_frame))) {
    printk(KERN_WARNING "wifi-emu: Received frame with incorrect length information\n");
    kfree(msg);
    return;
  }

  // allocate socket buffer
  skb = dev_alloc_skb(ntohs(msg->data_len));
  if(!skb) {
    printk(KERN_WARNING "wifi-emu: Could not allocate socket buffer\n");
    wifi_emu_mypriv->stats.rx_dropped += ntohs(msg->data_len);  
    kfree(msg);
    return;
  }

  // copy data
  memcpy(skb_put(skb, ntohs(msg->data_len)), msg->data, ntohs(msg->data_len));
  skb->dev = wifi_emu_mypriv->dev;
  skb->protocol = eth_type_trans(skb, wifi_emu_mypriv->dev);
  
  // update statistics
  wifi_emu_mypriv->stats.rx_packets++;  
  wifi_emu_mypriv->stats.rx_bytes += ntohs(msg->data_len);  
  
  kfree(msg);

  #if DEBUG_TIME
  do_gettimeofday(&t);
  printk(KERN_INFO "KERN_SendToOS\t%lu\t%06lu\n", t.tv_sec, t.tv_usec);
  #endif

  // submit frame
  // (netif_rx_ni calls netif_rx, but runs pending softirqs afterwards)
  // (otherwise the packet is not processed directly because we are in process context)
  netif_rx_ni(skb);

  return;
}

static void wifi_emu_recv_resp_scan(struct wifi_emu_msg_resp_scan *msg, int len)
{
  // check if already initalized
  if(wifi_emu_mypriv == NULL) { 
    printk(KERN_ERR "wifi-emu: Received message although not initialized");
    kfree(msg);
    return;
  }

  // check length
  if(msg->num_entries*sizeof(struct wifi_emu_scan_entry) > (len - sizeof(struct wifi_emu_msg_resp_scan))) {
    printk(KERN_WARNING "wifi-emu: Received scan information with incorrect length information\n");
    kfree(msg);
    return;
  }

  mutex_lock(&wifi_emu_scan_mutex);

  // free old result if needed
  if(wifi_emu_scan != NULL) {
    kfree(wifi_emu_scan);
  }

  // store this message (wifi_emu_get_scan will look for it)
  wifi_emu_scan = msg;

  mutex_unlock(&wifi_emu_scan_mutex);

  // do not free (msg is still used)
  return;
}

static void wifi_emu_recv_resp_fer(struct wifi_emu_msg_resp_fer *msg)
{
  // check if already initalized
  if(wifi_emu_mypriv == NULL) { 
    printk(KERN_ERR "wifi-emu: Received message although not initialized");
    kfree(msg);
    return;
  }

  // get_fer method is waiting for wifi_emu_fer to become != NULL
  // (we pass the message received with it)
  mutex_lock(&wifi_emu_fer_mutex);
  wifi_emu_fer = msg;
  mutex_unlock(&wifi_emu_fer_mutex);
  wake_up(&wifi_emu_wq_resp);

  // do not free (wifi_emu_fer points to the message)
  return;

}

static void wifi_emu_recv_spy_update(struct wifi_emu_msg_spy_update *msg)
{
  struct iw_quality qual;

  // check if already initalized
  if(wifi_emu_mypriv == NULL) { 
    printk(KERN_ERR "wifi-emu: Received message although not initialized");
    kfree(msg);
    return;
  }
  
  qual.qual = msg->qual.qual;
  qual.level = msg->qual.level;
  qual.noise = msg->qual.noise;
  qual.updated = IW_QUAL_ALL_UPDATED | IW_QUAL_DBM;

  wireless_spy_update(wifi_emu_mypriv->dev, msg->addr, &qual);

  kfree(msg);
  return;
}

static void do_update_stats(void)
{
  // update device type
  if(wifi_emu_mypriv->wstats.mode == WIFI_EMU_MODE_MONITOR)
    wifi_emu_mypriv->dev->type = ARPHRD_IEEE80211_RADIOTAP;
  else
    wifi_emu_mypriv->dev->type = ARPHRD_ETHER;
}


EXPORT_SYMBOL(wifi_emu_init);
EXPORT_SYMBOL(wifi_emu_cleanup);
EXPORT_SYMBOL(wifi_emu_recv_msg);


